'''pallete.py

This represents the twelve hues (RYGCBM, OLTSPV) across the two tones (saturated
and dark). The two other tones (light, unsaturated) are 'reserved' by convention
but not represented here.

We use capital letters and lowercase letters to refer to saturated and dark,
respectively. (e.g. pR = ord('R'), corresponding to '#ff0000', and
pr = ord('r'), corresponding to '#800000'.

This code provides useful enums, etc. With no imports, an
'import palette from *' should be fine. (Note that 'from package import *' is
usually a bad idea!)

If you want to import all the enums from this namespace and want to avoid *,
but you *also* don't want to write 'palette.pR' (because it's so long!)
then here are some easy copy-pastes for you:

from palette import pR, pY, pG, pC, pB, pM
from palette import pO, pL, pT, pS, pP, pV
from palette import pr, py, pg, pc, pb, pm
from palette import po, pl, pt, ps, pp, pv
from palette import get, resel_to_rgb, rgb_to_resel

'''

# palette.py
# These represent the twelve hues, across two tones (saturated, dark) from the palette in README.md
# By convention, 

# Palette stuff below!
# These are useful enums for colors in our palette.
# Here, we assign enums to match meaningful characters rather than arbitrary ints.
# (where 'meaningful character' is taken from enums.)

# These letters correspond to RYGCBM OLTSPV
# (Red, Yellow, Green, Cyan, Blue, Magenta; Orange, Lime, Teal, Sapphire, Purple, Violet)

pR = ord('R')   # 82    # On red wire
pr = ord('r')   # 114   # Off red wire
pG = ord('G')   # 71    # Unused
pg = ord('g')   # 103   # Unused
pB = ord('B')   # 66    # On blue wire
pb = ord('b')   # 98    # Off blue wire

pY = ord('Y')   # 89    # Unused
py = ord('y')   # 121   # Unused
pC = ord('C')   # 67    # XOR node
pc = ord('c')   # 99    # AND node
pM = ord('M')   # 77    # Output node (from input/logic node to wire)
pm = ord('m')   # 109   # Input node  (from wire to input/logic node)

pO = ord('O')   # 79
po = ord('o')   # 111
pL = ord('L')   # 76
pl = ord('l')   # 108
pT = ord('T')   # 84
pt = ord('t')   # 116
pS = ord('S')   # 83
ps = ord('s')   # 115
pP = ord('P')   # 80
pp = ord('p')   # 112
pV = ord('V')   # 86
pv = ord('v')   # 118

#_pal is just a helper; we loop over it to create our resel-rgb dicts.
# This represents the 12 hues, 'Saturated' and 'Dark'
_pal = [
    (pR, (255,0,0)),
    (pr, (128,0,0)),
    (pG, (0,255,0)),
    (pg, (0,128,0)),
    (pB, (0,0,255)),
    (pb, (0,0,128)),
    (pY, (255,255,0)),
    (py, (128,128,0)),
    (pC, (0,255,255)),
    (pc, (0,128,128)), 
    (pM, (255,0,255)),
    (pm, (128,0,128)),
    (pO, (255,128,0)),
    (po, (128,64,0)),
    (pL, (128,255,0)),
    (pl, (64,128,0)),
    (pT, (0,255,128)),
    (pt, (0,128,64)),
    (pS, (0,128,255)),
    (ps, (0,64,128)),
    (pP, (128,0,255)),
    (pp, (64,0,128)),
    (pV, (255,0,128)),
    (pv, (128,0,64)),
]

# We implement a cheap bidict here
resel_to_rgb = dict()
rgb_to_resel = dict()

for resel, rgb in _pal:
    rgb_to_resel[rgb] = resel
    resel_to_rgb[resel] = rgb

# And we implement a cheap defaultdict with this function
def get(dd, kk, default=0):
    """Get dd[kk], or return default if kk is not in dd.keys()
    
    :param dd: Dictionary
    :type dd: dict
    :param kk: Any hashable key
    :type kk: object
    :param default: The default value to return if kk is not a valid key
    :type default: object (anything!)
    
    :return: dd[kk], or 'default' if kk is not in dd.keys
    :rtype: object (anything!)
    """
    if kk in dd.keys():
        return dd[kk]
    else:
        return default




import argparse
from math import log, ceil
from time import time
import numpy as np
from PIL import Image
from .resoboard import ResoBoard

def main(
    load_filename,
    save_prefix,
    iterations = 1,
    save_each_iteration = True,
    V = False):
    """Wraps the logic in 'resoboard' for simulating, exporting, etc.
    
    Used in the actual '__main__' of this function, with parameters passed from
    argparser.
    
    :param load_filename: location from which to load the image from
    :type load_filename: String
    :param save_prefix: location to save the file to.
    :type save_prefix: String
    :param iterations: Number of simulation steps to update the circuit.
    :type iterations: Int
    :param save_each_iteration: If true, save an image of the circuit each iteration.
    :type save_each_iteration: Bool
    :param V: If True, print verbose output while running
    :type V: Bool
    """
    
    # See this ugly variable here?
    # This is why Python gave us fstrings.
    # todo: Use fstring magic and see if things still print right.
    num_digits_in_fname = ceil(log(iterations+.1,10))
    
    if V:
        print("  Loading", load_filename,"and iterating",iterations,"time(s),")
        print("  and then saving to ",save_prefix,'x'*num_digits_in_fname,".png",sep='')
    
    # Instantiate our ResoBoard
    compile_start = time()
    RB = ResoBoard(load_filename)
    compile_end = time()
    
    if V:
        print(f"... Compiled in {compile_end - compile_start:.2f} seconds! Iterating now.")
    
    # Simulation!
    iter_start = time()
    for ii in range(iterations):
        if save_each_iteration:
            save_loc = save_prefix + str(ii).zfill(num_digits_in_fname) + ".png"
            Image.fromarray(np.swapaxes(RB.get_image(),0,1)).save(save_loc)
        if V:
            print("Iteration: ",ii)
        # update_image is true if we're on our last iteration.
        update_image = save_each_iteration or ii == iterations - 1
        RB.iterate(update_resels = False, update_image = update_image )
    
    iter_end = time()
    # Last iteration, always saved
    if V:
        print("Iteration: ", iterations)
        print(f"Completed {iterations + 1} steps in {iter_end - iter_start:.2f} seconds!")
    save_loc = save_prefix + str(iterations).zfill(num_digits_in_fname) + ".png"
    Image.fromarray(np.swapaxes(RB.get_image(),0,1)).save(save_loc)
    


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Reso - graphical circuit design cellular automata")
    parser.add_argument("load_location", help="Location to load image from.",
                        type=str, nargs=1)
    parser.add_argument("--save", "-s", help="Prefix to save images to.",
                        type=str, nargs=1)
    parser.add_argument("--numiter","-n",
                        help="iterate the reso board n times. Defaults to 1.",
                        type=int, nargs=1)
    parser.add_argument("--outputlast","-o",
                        help="Only save the final iteration of the board.",
                        action="store_true")
    parser.add_argument("--verbose","-v", help="Print extra information; useful for debugging.",
                        action="store_true")

    args = parser.parse_args()
    
    if args.load_location is None:
        raise ValueError
    
    if args.save is None:
        raise ValueError

    load_filename   = args.load_location[0]
    save_prefix     = args.save[0]
    iterations = 1 if args.numiter  is None else args.numiter[0]
    save_each_iteration = not args.outputlast
    V = args.verbose
    
    main(load_filename, save_prefix, iterations, save_each_iteration, V)



'''resoboard.py

A Reso circuit can be thought of as a 'board' that needs to be iterated upon.
This code does the heavy lifting! Some important concepts:

1. The 'palette' defines our base colors. We import a lot from 'palette'.
2. A 'Board' holds all the information for a Reso circuit.
   An 'image' is a grid of RGB pixels.
   A  'board' is a grid of Reso elements (called **resels**).
'''

import numpy as np
from PIL import Image

from .regionmapper import ortho_map, diag_map, RegionMapper
from .palette import pR, pY, pG, pC, pB, pM
from .palette import pr, py, pg, pc, pb, pm
from .palette import pO, pL, pT, pS, pP, pV
from .palette import po, pl, pt, ps, pp, pv
from .palette import get, resel_to_rgb, rgb_to_resel


# Wire and Node classes used below to hold data about the state during iteration
class Wire:
    """ A class representing a wire. It can be on or off (controlled by 'state'),
    and extra information is stored in 'next_state' during
    
    (This could be a dataclass! But I want to be backwards-compatible.)
    
    :param regionid: The integer representing the unique ID of the region that
        is this wire.
    :type regionid: Int
    :param state: Corresponds to the visually outward state of the wire
    :type state: Bool or Int
    :param next_state: Used only to calculate updates, and remains False between
        iterations.
    :type next_state: Bool or Int
    """
    def __init__(self, regionid, state = False, next_state = False):
        self.regionid = regionid
        self.state = state
        self.next_state = next_state


class Node:
    """Representing all other nodes, which are static and hold no other info.
    
    (This could be a dataclass! But I want to be backwards-compatible.)
    
    :param regionid: The integer representing the unique ID of the region that
        is this wire.
    :type regionid: Int
    :param state: CUsed only to calculate updates, and remains False between
        iterations.
    :type state: Bool or Int
    """
    # Node is everything that is not a wire
    def __init__(self, regionid, state = False):
        self.regionid = regionid
        self.state = state # Remains false between iterations
        # For 'and' nodes, state starts at False,
        # and state locks to -1 if connected to an off wire.


# todo: ResoBoard has wire colors hard-coded into it (pR, pr, pB, pb, etc.)
# I'd prefer, instead, to abstract all that away, so it's easier to play with
# other colors!
class ResoBoard:
    """Holds a compiled Reso circuit and performs the logic of simulating over
    said circuit.
    
    If you want to wrap around the core logic of Reso, this is what you want
    to use!
    
    (Sorry functional-programming fanatics, maintaining state allows the
    circuit to be compiled and operated on much more efficiently, and I don't
    know how to maintain state in FP here without the code being ugly as sin!)
    
    
    :param image: Numpy array of shape (w, h, 3) representing the RGB image.
    :type image: numpy.ndarray
    :param resel_to_rgb: Dict mapping 'resel' enums (per palette.py) to pixel
        values, i.e. RGB 3-tuples.
    :type resel_to_rgb: Dict
    :param rgb_to_resel: Dict mapping pixel values (i.e. RGB 3-tuples) to 'resel'
        enums.
    :type rgb_to_resel: Dict
    
    Note that resel_to_rgb and rgb_to_resel form a bidict, i.e.
        resel_to_rgb[rgb_to_resel[x]] = x, and
        rgb_to_resel[resel_to_rgb[x]] = x
    
    Member variables:
    _image: RGB image, numpy array, shape (w, h, 3)
    _resel_map: Grid of resel values, i.e. numpy array of shape (w, h)
    _RM: The RegionMapper object that actually maps regions of pixels/resels to 
        'regions'.
    _resel_objects: List of Wire() or Node() objects, indexed by their unique
        region IDs.
    _orange_wires, _sapphire_wires, _lime_wires
        Lists of Wire() objects, pointing to the same objects as in _resel_objects
     _inputs, _outputs, _ands, _xors:
        Lists of Node() objects, pointing to the same objects as in _resel_objects.
    
    A bunch of adjacency dicts:
    These are indexed by region_id, mapping to a list of all adjacent elements
    (that is, Wire() or Node() objects.)
    _adj_inputs: Indexed by ID of a wire element
    _adj_xors: Indexed by ID of an input node
    _adj_ands: Indexed by ID of an input node
    _adj_outputs: Indexed by ID of an input node, an XOR node, or an AND node.
    _adj_wires: Indexed by ID of output nodes.
    """
    def __init__(self,
        image,
        resel_to_rgb = resel_to_rgb,
        rgb_to_resel = rgb_to_resel
    ):
        """
        Initialization (1) Grabs the image, (2) converts it to self._resel_map,
        (3) identifies the different regions (wires, inputs, etc.) in
        self._resel map, (4) Sets up associations between those regions so that
        they can be quickly accessed, and (5) initializes Wire and Node objects
        corresponding to each region.
        
        
        """
        # First step: Load the image and convert it to _resel_map.
        # Here, the 'image' can be a string (which will be loaded)
        # or a prepared numpy array (of shape (w, h, 3).)
        if isinstance(image, str):
            image = np.swapaxes(np.array(Image.open(image)), 0, 1)[:,:,:3]
        # else: assume image is of format (width, height, 3), indexed (x,y)
        self._image = image
        
        # Now convert our image to a resel_map (e.g. (255,0,0) becomes pR).
        width = self._image.shape[0]
        height = self._image.shape[1]
        self._resel_map = np.zeros((width, height))
        # Yeah I'm nesting for-loops rather than vectorizing this through a GPU.
        # whatcha gonna do about it?!?
        # (msubmit a patch? please? i don't know how to do these things on GPU
        #  using Python, to be perfectly honest...)
        for ii in range(width):
            for jj in range(height):
                self._resel_map[ii,jj] = get(rgb_to_resel, tuple(self._image[ii,jj]))
        
        # Identify the different regions, giving us our self._RM (RegionMapper)
        # First, we note that 'on' wires and 'off' wires **are the same class!**
        # So, for our regionmapper, we need them to map to the same class.
        class_dict = { 
            pO : pO, po : pO,    # pO is also used to denote orange wires
            pS : pS, ps : pS,    # pS is also used to denote sapphire wires
            pL : pL, pl : pL,    # pL is also used to denote lime wires
            
            pP : pP, pp : pp,   # Every other color maps to itself
            pT : pT, pt : pt,
            
            pR : pR, pr : pr,
            pG : pG, pg : pg,
            pB : pB, pb : pb,
            pC : pC, pc : pc,
            pY : pY, py : py,
            pM : pM, pm : pm, 
            pV : pV, pv : pv
        }
        
        # Wires are diagonally contiguous (to make it easier for them to 'cross')
        # while everything else is only orthogonally continuous
        contiguities = {
            pO : ortho_map + diag_map,
            pS : ortho_map + diag_map,
            pL : ortho_map + diag_map
        } # Wires are diagonally contiguous
        
        # Now we use our RegionMapper helper to identify all the distinct,
        # contiguous regions that form the 'elements' of our circuit!
        self._RM = RegionMapper( self._resel_map,             
                                 class_dict     = class_dict,           
                                 contiguities   = contiguities,
                                 sparse         = True)
        # As a reminder, self._RM (RegionMapper) provides:
        #   self._RM.region_at_pixel(x,y)
        #   self._RM.regions(id)
        #   self._RM.regions_with_class(class)
        #   self._RM.adjacent_regions(region_id)
        
        # Now, we need to set up a Wire()/Node() object for each region.
        # We need lists of all such objects (self._orange_wires, etc...),
        # and the regionid -> Wire()/Node() list self._resel_objects.
        # This code is redundant in space, but we save time :)
        # (... todo: this can be cleaned up? We can just use
        # self._RM.regions_with_class(pO), right? It'll be uglier but who cares?)
        
        self._resel_objects = [None]*len(self._RM._regions)
        # # One entry in self._resel_objects for each region we have.
        self._orange_wires   = []
        self._sapphire_wires = []
        self._lime_wires     = []
        self._inputs         = []
        self._outputs        = []
        self._ands           = []
        self._xors           = []
        
        # Now we loop over all our regions, and their associated class type.
        for regionid, (classid, _) in enumerate(self._RM._regions):
            if classid == pO or classid == pS or classid == pL:
                # Recall that off wires and on wires were both mapped to the same class.
                new_object = Wire(regionid)
                if classid == pO:
                    self._orange_wires.append(new_object)
                elif classid == pS:
                    self._sapphire_wires.append(new_object)
                elif classid == pL:
                    self._lime_wires.append(new_object)
                else:
                    raise ValueError('Check wire mapping loop. This shouldn\'t be possible to see!')
            else:
                # It's not a wire, we have a node instead!
                # We need to add it to the correct list of elements.
                new_object = Node(regionid)
                if classid == pp:
                    self._inputs.append(new_object)
                elif classid == pP:
                    self._outputs.append(new_object)
                elif classid == pt:
                    self._ands.append(new_object)
                elif classid == pT:
                    self._xors.append(new_object)
                elif classid in (
                    pR, pr, pG, pg, pB, pb, pC, pc, pY, py, pM, pm, pV, pv
                ):
                    # List of all reserved but not-yet-used colors
                    pass
                else:
                    # This should never happen!
                    # Our region mapper should not have been able to get to this state...
                    print("Unrecognized classid", classid)
                    raise ValueError('Somehow, we mapped a class outside the palette. Shouldn\'t be possible!')
            
            # We also want whatever object we create to be stored in our list
            # of all _resel_objects.
            # (todo... can't we, at the end of this loop, add all our other lists
            #  together to make one big list?)
            self._resel_objects[regionid] = new_object
        
        # region), then that whole wire should be considered on.
        # So, loop over every wire, and if any pixel is 'on', then turn it on!
        for wires in (self._orange_wires, self._sapphire_wires, self._lime_wires): 
            for wire in wires:
                wire_is_on = False
                pixels_in_wire = self._RM.regions(wire.regionid)[1]
                for ii, jj in pixels_in_wire:
                    if self._resel_map[ii, jj] in (pO, pS, pL):
                        wire_is_on = True
                        break
                wire.state = wire_is_on

        # Now we set up adjacency dictionaries.
        # For each of these, dict[region_id] -> Wire()/Node() object
        # _adj_inputs[]  for wire_id, 
        # _adj_xors[]    for all input cells
        # _adj_ands[]    for all input cells
        # _adj_outputs[] for all input cells and all logic nodes
        # _adj_wires[]   for all output cells
        
        self._adj_inputs    = dict()
        self._adj_xors      = dict()
        self._adj_ands      = dict()
        self._adj_outputs   = dict()
        self._adj_wires     = dict()

        # Loop over the following:
        # For every region in _orange_wires, create an entry in self._adj_xors[region]
        # for every *adjacent* region having a class in classids
        for from_list, to_dict, classids in \
            [((self._orange_wires + self._sapphire_wires + self._lime_wires),
              self._adj_inputs, (pp,)),
             (self._inputs, self._adj_xors, (pT,)),
             (self._inputs, self._adj_ands, (pt,)),
             (self._inputs, self._adj_outputs, (pP,)),
             (self._xors, self._adj_outputs, (pP,)),
             (self._ands, self._adj_outputs, (pP,)),
             (self._outputs, self._adj_wires, (pO, pS, pL))]:
             
             for resel in from_list:
                to_dict[resel.regionid] = []
                for adj_reg_id in self._RM.adjacent_regions(resel.regionid):
                    adj_reg_class = self._RM.regions(adj_reg_id)[0]
                    if adj_reg_class in classids:
                        to_dict[resel.regionid].append(self._resel_objects[adj_reg_id])
        
        # Finally,  we want our cheap bidict for converting resels to pixels
        # and  vice-versa
        self.rgb_to_resel = rgb_to_resel
        self.resel_to_rgb = resel_to_rgb
    
    
    def _update(self, resel_map = False, update_image = True):
        """Update the values in resel map and in the image.
        This updates the "externally visible" parts of a ResoBoard.
        
        For speed, this isn't necessary to do every time! None of the information
        updated in this function is necessary for the logic, only the output.
                
        :param resel_map: If True, update our _resel_map of classes
        :type resel_map: bool
        :param image: If True, update our RGB _image
        :type image: bool
        """
        # Only loop if we have something we want to update!
        if resel_map or update_image:
            for oncolor, offcolor, wires in (
                (pO, po, self._orange_wires),
                (pS, ps, self._sapphire_wires),
                (pL, pl, self._lime_wires)
            ):
                # "oncolor" refers to 'saturated red' and 'saturated blue'
                # "offcolor" refers to 'dark red' and 'dark blue'
                # and 'wires' are the list of all wires that we have
                for wire in wires:
                    # Get the RGB color (color_tuple) the set the pixel to
                    color = oncolor if wire.state else offcolor
                    color_tuple = np.array(self.resel_to_rgb[color])
                    # Get every pixel in the region
                    # (RegionMapper.regions(regionid)[1] is the list of pixels)
                    regionid = wire.regionid
                    pixel_list = self._RM.regions(regionid)[1]
                    
                    for ii, jj in pixel_list:
                        if resel_map:
                            # 'color' is one of pO, po, pS, ps, pL, pl
                            self._resel_map[ii,jj] = color
                        if update_image:
                            # 'color_tuple' is the RGB tuple
                            self._image[ii,jj] = color_tuple
                
    
    def get_resel_map(self):
        """Return the Numpy array representing the Reso board
        
        :returns: The [w,h] Numpy array representing the Reso board
        :rtype: numpy.ndarray
        """
        # 
        return self._resel_map
    
    def get_image(self):
        """Return the Numpy array containing the underlying image.
        
        :returns: The [w,h,3] Numpy array containing the underlying image.
        :rtype: numpy.ndarray
        """
        return self._image
    
    def iterate(self, update_resels = True, update_image = True):
        """Iterate the board, updating every Wire() object.
        This is the 'main logic' of updating a Reso circuit.
        
        Also updates the resels if update_resels is True,
        and also updates the image if update_image is True
        
        Short description of the steps:
        1. Push wire values through input nodes to logic nodes!
        2. Push logic nodes values through to output nodes!
        3. Push output nodes values to wires!
        
        Longer description:
        1. Every logic node stores an internal boolean value which is updated
           according to the values of wires of adjacent input nodes.
           Output nodes are updated as if they were logical 'or' nodes.
           ('xor' starts at False, and is flipped for each 'True' input.)
           ('and' starts at False, is flipped 'True' if any 'True' inputs are
            seen, and is permanently set to -1 (for 'False') if any 'False'
            inputs are seen.)
           ('output' nodes connected to an input node starts at False, and are
            updated to True if any 'True' inputs are seen. I.e. an 'or')
        2. Every logic node pushes their value to adjacent output nodes. This is
           done as a logical 'or', i.e. if any adjacent logic nodes have a
           'True' value, then the output node is set true!
        3. Every output node then pushes its logical value to wires, using a
           logical 'or' just as in step 2.
                
        :param resel_map: If True, update our _resel_map of classes
        :type resel_map: bool
        :param image: If True, update our RGB _image
        :type image: bool
        """
        
        # Update all the 'input' nodes connected to wires
        # and then update every connected logic node ('xor', 'and')
        for wire in (self._orange_wires + self._sapphire_wires + self._lime_wires):
            # For each input connected to that wire,
            for inputnode in self._adj_inputs[wire.regionid]:
                # Update the internal states of adjacent xor, and, outputs
                for xornode in self._adj_xors[inputnode.regionid]:
                    xornode.state = xornode.state ^ wire.state
                
                for andnode in self._adj_ands[inputnode.regionid]:
                    if not (andnode.state == -1):
                        if wire.state == False:
                            andnode.state = -1
                        else:
                            andnode.state = True
                for outnode in self._adj_outputs[inputnode.regionid]:
                    outnode.state = outnode.state or wire.state
        
        # For each 'xor' logical element,
        # update the state of connected output nodes
        for xornode in self._xors:
            for outnode in self._adj_outputs[xornode.regionid]:
                outnode.state = outnode.state or xornode.state
        
        # For each 'and' logical element,
        # update the state of connected output nodes
        for andnode in self._ands:
            if andnode.state == True:
                for outnode in self._adj_outputs[andnode.regionid]:
                    outnode.state = outnode.state or andnode.state
        
        # For each output node, update the state of connected wires.
        for outnode in self._outputs:
            # Update the next_state of each adjacent wire
            for wire in self._adj_wires[outnode.regionid]:
                 wire.next_state = wire.next_state or outnode.state
        
        # Finally, reset the states of every wire.
        # We used 'next_state' just as a placeholder during iteration
        for wire in (self._orange_wires + self._sapphire_wires + self._lime_wires):
            wire.state = wire.next_state
            wire.next_state = False

        for node in self._xors + self._ands + self._inputs + self._outputs:
            node.state = False
        
        # By default, also updates the resels and the image
        self._update(update_resels, update_image)



import numpy as np

# todo:
# consider making this faster,
# per https://stackoverflow.com/questions/12321899/
# this is used in the 'compilation' step

# todo: this big docstring isn't necessary...

'''
Provides:

ortho_map = ((1,0),  (0,-1), (-1,0), (0,1))
diag_map  = ((1,-1), (-1,-1),(-1,1), (1,1))

    These are used to calculate adjacent pixels.
    Useful for the contiguities and adjacencies dictionaries for RegionMapper

RegionMapper(image,
             class_dict,
             contiguities = {},
             adjacencies = {},
             sparse = True,
             wrap = False)
    Parameters:
        image: A Numpy array of shape (width, height, n_channels)
            E.g. image[x,y] = [255, 0, 0]
                Note: Axis 0 and 1 should represent x and y, respectively!
                If they're wrong, use np.swapaxes(picture, 0, 1)
        class_dict: A dictionary mapping tuple of size n_channels --> int
            E.g. class_dict = { (255,0,0):1, (0,255,0):2, (0,0,255):3 }
                # Red is class 1, green is class 2, blue is class 3
        contiguities: A dictionary mapping classes to what they consider 'contiguous regions'
            E.g. contiguities = { 1 : ortho_map + diag_map, 3: diag_map }
                # Red regions are contiguous if they touch orthogonally or diagonally
                # Blue regions are contiguous only if they touch diagonally
                # Green regions are orthogonally contiguous (default)
        adjacencies: Same format as contiguities. Defines what is considered a neighbour for a given class.
            E.g. adjacencies = {} means (by default) regions of any class will only consider
            other regions to be its neighbour if a pixel of that region is orthogonally adjacent
        sparse: Boolean. Set True if most of the pixels in the image do not belong to a given class.
        wrap: Boolean. If True, the image is considered a torus. The top edge is adjacent to the bottom edge,
            and the left edge is adjacent to the right edge.

    Provides:
    RegionMapper.region_at_pixel(x,y):
        Returns the ID of the region at that pixel, or
        -1 if that region does not belong to a class.
    RegionMapper.regions(region_id):
        Given the ID of a region, return (class_number, list of pixels in that region)
    RegionMapper.regions_with_class(class_number):
        Given the number of a class, return all regions with that class.
    RegionMapper.adjacent_regions(region_id):
        Given the ID of a region, return all regions adjacent to it (as defined by
        its entry in the adjacencies dictionary)
'''


# Package constants, useful for defining neighborhoods.
# ortho_map defines an orthogonal neighborhood
# diag_map  defines a diagonal neighborhood
ortho_map = ((1,0),  (0,-1), (-1,0), (0,1))
diag_map  = ((1,-1), (-1,-1),(-1,1), (1,1))

def _value_to_class(class_dict, value):
    """ Map numpy array values using a dictionary.
    Convenience function for dictionaries that are keyed on tuples or integers.

    Used to map integers (numpy arrays of integers, shape (3,)) to integers.
    These integers enumerate the class that a pixel belongs to.

    :param class_dict: A mapping of tuples of integers (i.e. RGB pixels) to integers
    :type class_dict: dict
    :param value: The key used to index the class_dict.
    :type value: numpy.ndarray

    :raises ValueError: If the value has a shape with 2 or more axes.

    :returns: A mapping per class_dict, or 0 if that value is not available.
    :rtype: Anything

    >>> _value_to_class(
    ...     {(1,2,3) : 'some_class', 789 : 'another_class'},
    ...      np.array([1,2,3]))
    'some_class'

    >>> _value_to_class(
    ...     {(1,2,3) : 'some_class', 789 : 'another_class'},
    ...      np.array(789))
    'another_class'
    """
    # This was written when I was a baby. There's probably a better way to do
    # this using dictionaries.

    # E.g. If the shape of the array = (3,), e.g. np.array([255, 255, 255])
    if len(value.shape) == 1:
        if tuple(value) in class_dict.keys():
            return class_dict[tuple(value)]
        else:
            return 0
    # E.g. if the shape of the array = (), e.g. np.array(12345)
    elif len(value.shape) == 0:
        if value in class_dict.keys():
            return class_dict[value]
        else:
            return 0
    else:
        raise ValueError


def _class_to_map(nbhd_offsets, value, default=ortho_map):
    """Index a dictionary 'nbhd_offsets' on 'value', returning 'default' if
    'value' is not a valid key.

    :param nbhd_offsets:
    :type nbhd_offsets: dict
    :param value: The key used to index the dictionary 'nbhd_offsets'
    :type value: Any hashable object
    :param default: Default value to return if value is not in the keys, defaults
        to regionmapper.ortho_map
    :type default: Any object

    :returns: todo
    :rtype:
    """
    # e.g. _class_to_map(contiguities, 3)
    # where contiguities = { 3 : ortho_map + diag_map, ... }
    # returns contiguities[3]
    # This is used to identify the x and y offsets used in the contiguities and adjacencies

    # This was writen when I was a baby.
    # This can probably be replaced with collections.defaultdict... todo!

    if value in nbhd_offsets.keys():
        return nbhd_offsets[value]
    else:
        return default


def _get_adjacent_pixels(x, y, w, h, nbhd_map = ortho_map, wrap = False):
    """Returns a list of indices for adjacent pixels given a 'neighborhood map'.

    Where (x,y) is the central pixel index, this will return a list of adjacent
    pixels, as defined by the offsets in 'nbhd_map'.

    E.g. for (x,y) = (3,6) and nbhd_map = ((1,0),  (0,-1), (-1,0), (0,1)),
    this will return [(4, 6), (3, 5), (2, 6), (3, 7)].

    The 'wrap' parameter is only important if (x,y) are near the (w,h) boundaries.
    See the examples

    :param x: X-coordinate of the image
    :type x: Int
    :param y: Y-coordinate of the image
    :type y: Int
    :param w: Width of the image
    :type w: Int
    :param h: Height of the image
    :type h: Int
    :param nbhd_map: Iterable of tuple defining the "neighborhood". Defaults to
        region_mapper.ortho_map.
    :type nbhd_map:
    :param wrap: If True, wrap around the dictionary.
    :type wrap: Bool

    :returns: List of indices representing 'neighbors' of a pixel.
    :rtype: list of tuple of int

    >>> _get_adjacent_pixels(3,6,10,10)
    [(4, 6), (3, 5), (2, 6), (3, 7)]

    >>> _get_adjacent_pixels(x=3,y=6,w=4,h=7)
    [(3, 5), (2, 6)]

    >>> _get_adjacent_pixels(x=3,y=6,w=4,h=7, wrap=True)
    [(0, 6), (3, 5), (2, 6), (3, 0)]
    """
    # Returns only valid pixels; order of pixels not guaranteed
    # E.g. adjacent_pixels(0,0,3,4) = [(1,0),(0,3),(0,2),(0,1)]
    # E.g. adjacent_pixles(0,0,0,0, wrap=False) = [(1,0),(0,1)]
    adj_pixels = []
    for offset_x, offset_y in nbhd_map:
        if wrap:
            adj_pixels.append(((x+offset_x) % w, (y+offset_y) % h))
        else:
            if 0 <= x + offset_x < w and 0 <= y + offset_y < h:
                adj_pixels.append((x+offset_x, y + offset_y))

    return adj_pixels


class RegionMapper:
    """
    TODO -- document this, the init function, and clean up the comments

    Given an image, the goal is to identify contiguous regions of the same color
    pixel, where 'contiguity' is defined per-color.

    A small glossary:
     - Pixel: The basic element used in the region-mapper. Each pixel has colors
        defined as 3 ints (e.g. such as (255,0,127)) and an (x,y) coordinate.
     - Coordinate: Or pixel index, i.e. the (x,y) coordinate that defines a
        unique location in an image.
     - Class: An integer associated with the color of a pixel.
        E.g. One might have {(255,0,0) : 1, (0, 255, 0) : 2, (0, 0, 255) : 3},
        i.e. associated red, green, and blue with integers 1, 2, and 3.
     - Contiguity: Contiguity is defined per-class. For example, red pixels might
        only be contiguous on orthogonal neighbors, while blue pixels might be
        contiguous on both orthogonal and diagonal neighbors.
        "Contiguity" is defined per-class for flexibility.
     - Region: A collection of pixels sharing the same class, defined by their
        'contiguity'. Every region has a unique integer 'region id'.
     - Region ID: A unique integer associated with each region after the
        region-mapping algorithm is finished.
     - Adjacency: A region that is next to another region. Adjacencies to regions
        are defined similarly to contiguities.


    :param image: A numpy array of integers (0-255) in shape (w, h, 3)
    :type image: numpy.ndarray
    :param class_dict: Dictionary mapping 3-tuples (pixel) to ints (class)
    :type class_dict: Dict
    :param contiguities: Dictionary of class int --> list of coordinates.
        (E.g. One might use 'ortho_map' here.)
    :type contiguities: Dict
    :param adjacencies: Dictionary of class int --> list of coordinates.
        (E.g. One might use 'ortho_map' here.)
    :type adjacencies: Dict
    :param sparse: If True, use a dictionary instead of an array to map pixels to regions.
        (This is useful if most of the pixels in the image don't correspond to
        a class.)
    :type sparse: Bool
    :param wrap: If True, region adjacencies wrap around the edge of the image.
        (Like a torus, or teleporting through the side of the screen like in Pacman.)
    :type wrap: Bool
    """
    def __init__(self,
                 image,                 # 2D Numpy Array
                 class_dict,            # Dict of (value) --> Int >= 1
                 contiguities   = {},   # Dict of class int --> map (like ortho_map)
                 adjacencies    = {},   # Dict of class int --> map (like ortho map)
                 sparse         = True,
                 wrap           = False):

        # TODO: Split this off into other functions!
        width, height = image.shape[:2]

        assert(len(image.shape) == 3 or len(image.shape) == 2), \
            "image should be np array shaped (width, height) or (width, height, number_of_channels_in_image)"

        # Create self._image, holding a 2D numpy array of class ints
        # I.e. Convert an rgb-image (w,h,3) to class-image (w,h)
        # todo: this can be vectorized !!!
        self._image = np.zeros((width, height))
        for ii in range(width):
            for jj in range(height):
                self._image[ii, jj] = _value_to_class(class_dict, image[ii,jj])

        # Mapping of pixel coordinates to region indices.
        # If Sparse, self._region_at_pixel this is a dict mapping (x, y) tuples to int
        # otherwise, self._region_at_pixel is an *array* mapping [x,y] to int
        # Where the 'int' is a unique integer mapping to the region
        # An ID of -1 means there is no region at this pixel
        self._regions = []
        if sparse:
            self._region_at_pixel = dict()
        else:
            self._region_at_pixel = np.zeros((width, height)) - 1

        # _regions_with_class:
        # E.g. _regions_with_class[2] = [1,3,4] # Regions 1, 3, and 4 are the ones with class 3
        # Basically, we also want a mapping of all regions that have a certain class.
        self._regions_with_class = dict()

        # The busy work!
        # 'Region' integer IDs are incremented from 0 up
        region = 0
        # if rec[x,y] == True, then this pixel has already been recorded.
        rec = np.zeros((width, height), dtype=np.bool)
        for x in range(width):
         for y in range(height):
                # Each pixel already has its classed assigned,
                # where 0 indicates no class
                region_class = self._image[x,y]
                if rec[x,y] == False and not region_class == 0:
                    # An unexplored region with a class!
                    # Let's explore it, and list all the pixels in it.
                    list_of_pixels_in_region = []

                    # Here, we fill up Fill up list_of_pixels_in_region using BFS
                    # contig_map is A list of what we consider "contigous" pixels.
                    contig_map = _class_to_map(nbhd_offsets = contiguities, value = region_class)
                    # pixels_under_consideration:
                    # Every pixel we are looking at but have not computed yet.
                    pixels_under_consideration = [(x,y)]
                    rec[x,y] = True

                    while len(pixels_under_consideration) > 0:
                        # Loop through all our pixels. (BFS starts here.)
                        # Takes the first element from pixels_under_consideration...
                        xi, yi = pixels_under_consideration.pop()
                        list_of_pixels_in_region.append((xi, yi)) # = [..., (x,y)]
                        self._region_at_pixel[xi, yi] = region
                        # For each adjacent pixel,
                        #     if it's the same class,
                        #     add it to the list of pixels we're going to explore
                        for xJ, yJ in _get_adjacent_pixels(
                            x=xi, y=yi, w=width, h=height, nbhd_map = contig_map, wrap = wrap
                        ):
                            if self._image[xJ, yJ] == region_class and rec[xJ, yJ] == False:
                                # if the pixel is inside the region of our origin pixel,
                                # and if it is not already recorded...
                                pixels_under_consideration.append((xJ, yJ))
                                rec[xJ, yJ] = True

                    # we're done! add to our regions our
                    # (region_class, pixel_pixel_pixel ...)
                    self._regions.append((region_class, list_of_pixels_in_region))

                    # Add to our mapping of class --> region ID
                    # (and create the entry for the class index first,
                    #  if this is our first time seeing it.)
                    if not (region_class) in self._regions_with_class.keys():
                        self._regions_with_class[region_class] = [region]
                    else:
                        self._regions_with_class[region_class].append(region)

                    # finally, increment our region index
                    region += 1

        # We did it, we mapped all our regions!
        # Now it's time to identify adjacent regions.
        #     self._adjacent_regions[region_id] provides a list of region_ids
        #     of adjacent regions, as defined by the 'adjacencies'
        #     associated with that class.
        self._adjacent_regions = [[] for _ in range(len(self._regions))]

        # ii is region id
        for ii, (region_class, list_of_pixels) in enumerate(self._regions):
            # nbhd_map = The list of pixel offsets of what this class considers its neighbours
            nbhd_map = _class_to_map(nbhd_offsets = adjacencies, value = region_class)

            # For every pixel in region ii,
            for xi, yi in list_of_pixels:
                # For every pixel adjacent to that pixel,
                for xJ, yJ in _get_adjacent_pixels(
                    x=xi, y=yi, w=width, h=height, nbhd_map=nbhd_map, wrap=wrap
                ):
                    # If the neighbour is a valid region (not empty)
                    if not self._image[xJ,yJ] == 0:
                        neighbour = self._region_at_pixel[xJ,yJ]
                        if not (neighbour == ii) and \
                            not neighbour in self._adjacent_regions[ii]:
                            # If the neighbour is not us
                            #    and it's not the same class as us
                            #    and if we didn't already consider this neighbour...
                            # (neighbour is the integer ID for that region...)
                            # Add that neighbour to our list of regions!
                            self._adjacent_regions[ii].append(neighbour)


    # Helper functions from here on.
    def region_at_pixel(self, x, y):
        """Returns the ID of the region at that pixel,
            or -1 if that region does not belong to a class.
        :param x: x-index of pixel contained in the Region Mapper
        :type x: Int
        :param y: y-index of pixel contained in the Region Mapper
        :type y: Int

        :returns: The class at pixel (x,y), or -1 if such a class does not exist.
        :rtype: Int
        """
        # todo: What was I thinking with this design?
        # What was I thinking in 2018?
        # If you see this, don't judge me~
        if self._image[x,y] == 0:
            return -1
        else:
            return self._region_at_pixel[x,y]


    def regions(self, region_id):
        """Given the ID of a region, return
        (class_number, list of pixels indices in that region),
        i.e. a list of (x,y) coordinates.

        E.g. regions(region_at_pixel(x=10,y=12)) will tell you (1) the type of
        region at (10,12), and (2) every other pixel at that region.

        :param region_id: Integer that maps to a region.
        :type region_id: Int

        :returns: Tuple of (class) and (list of pixel indices) in that region.
        :rtype: Tuple
        """
        return self._regions[region_id]


    def regions_with_class(self, class_number):
        """Given the number of a class, return all regions with that class.

        For a class with integer 3, return a list of all region indices
        that have that class index.

        To get a list of pixel indices for a given region index,

        :param class_number: The integer associated with a class.
        :type class_number: Integer

        :returns: List of region IDs that have a given class number.
        :rtype: List of int
        """
        return self._regions_with_class[class_number]

    def adjacent_regions(self,region_id):
        """ Given the ID of a region, return all regions adjacent to it
        (as defined by its entry in the adjacencies dictionary)

        :param region_id: The ID associated with a region of pixels.
        :type region_id: Int

        :returns: List of all region IDs of regions adjacent to a given region ID.
        :rtype: List of int (or empty list)
        """
        return self._adjacent_regions[region_id]
